// Lilo.cpp
//
// Copyright (c) 1998 Otfried Cheong.
//
#include <k32std.h>
#include <e32std.h>
#include "arlo.h"

// version information
const TInt KMajorVersionNumber=0;
const TInt KMinorVersionNumber=4;
const TInt KBuildVersionNumber=4;

// driver definition

class DArLoDriver : public DLogicalDevice
{
public:
  DArLoDriver();
  virtual TInt Install();
  virtual void GetCaps(TDes8 &aDes) const;
  virtual DLogicalChannel *CreateL();
  virtual ~DArLoDriver();
};

// channel definition

class DArLo : public DLogicalChannel
{
public:
  DArLo(DLogicalDevice *aDevice);
protected:
  virtual void DoCancel(TInt aReqNo);
  virtual void DoRequest(TInt aReqNo,TAny *a1,TAny *a2);
  virtual TInt DoControl(TInt aFunction,TAny *a1,TAny *a2);
};

/*
  class DLogicalDevice implementation
*/

// a .LDD requires that the first exported function create the device.
// By returning a concrete DLogicalDevice-derived class, this implements
// polymorphism using a named DLL

EXPORT_C DLogicalDevice *CreateLogicalDevice()
{
  return new DArLoDriver; // this is the type of device we return
  // if the new failed, and returned 0, then the user program which
  // attempts to load the device will fail.  We do not need to take any
  // specific action here to handle out-of-memory
}

DArLoDriver::DArLoDriver()
{
  iVersion=TVersion(KMajorVersionNumber,
		    KMinorVersionNumber,
		    KBuildVersionNumber);
  // just set the version
}

TInt DArLoDriver::Install()
{
  TPtrC name=_L("ArLo");       // our name, used by User::FreeLogicalDevice
  return SetName(&name);       // set our name and return error code
}

DArLoDriver::~DArLoDriver()
{ }

void DArLoDriver::GetCaps(TDes8 &aDes) const
{
  // standard capability-descriptor pattern
  TCapsArLoV1 b; // a structure
  b.version=iVersion; // fill it with information
  aDes.FillZ(aDes.MaxLength()); // zero-pad the caller's descriptor
  aDes.Copy((TUint8 *)&b,Min(aDes.MaxLength(),sizeof(b)));
  // copy as much of our information as will fit into this descriptor
}

DLogicalChannel *DArLoDriver::CreateL()
{
  // polymorphic function to create whatever kind of channel
  // the device is intended to support
  return new(ELeave) DArLo(this);
  // since we're a DArLoDriver, return a DArLo channel
}

/*
  class DArLo implementation - the channel
*/

DArLo::DArLo(DLogicalDevice *aDevice)
  : DLogicalChannel(aDevice)
{
  SetBehaviour(0); // indicate we do not support asynchronous requests
}

void DArLo::DoCancel(TInt /* aReqNo */)
{ // cancel any outstanding request
  // shouldn't happen!!
}

void DArLo::DoRequest(TInt aReqNo,TAny* /*a1*/,TAny* /*a2*/)
{ // kick off any asynchronous request
  Complete(aReqNo,KErrNotSupported); // request is complete (unhelpfully!)
}

#define __NAKED__ __declspec( naked )

__NAKED__ TInt DArLo::DoControl(TInt aFunction, TAny* a1, TAny* a2)
{
  // handle any synchronous request function
  // can trash r0-r3 and ip
  // on entry:
  //   r0=this
  //   r1=aFunction
  //   r2=a1 (PhysMapAdr)
  //   r3=a2 (BootAdr)
  // on exit:
  //   r0=result
  //                                  

  asm("  cmp r1,#0");
  asm("  beq boot");

  // Examine an address
  // [R2] contains address to examine
  asm("  ldr r3,[r2]"); // get address to examine in R3
  asm("  ldr r1,[r3]"); // get value
  asm("  str r1,[r2]"); // store value back into [R2]
  asm("  mov  r0,#0");  // no error
  asm("  mov  pc,lr");  // return

  // Booting
  asm("boot: ");

  // turn off interrupts
  asm("  mrs r0, cpsr");
  asm("  bic r0, r0, #0xdf");
  asm("  orr r0, r0, #0xd3");  // disable FIQ, IRQ, set SVC mode
  asm("  msr cpsr,r0");

  // load r0 with physical map address
  asm("  mov r0, r2");

  // turn off MMU (write zero into MMU register)
  asm("  mov R1,#070");      // not quite zero---don't want 26 bit mode
  asm("  mcr p15,0,R1,c1,c1");

  // we have two instructions here before hell breaks lose!
  // jump to R3
  asm("  mov pc,R3");
  // will never get here
}

// DLL entry point, required by all DLLs,
// called on thread or process attach/detach
GLDEF_C TInt E32Dll(TDllReason /*aReason*/)
{
  return KErrNone; // do nothing, return no error
}

